# -*- encoding: utf-8 -*-
# Module iainfrec

import numpy as np
from iasecross import iasecross
import ia870 as MT

def iainfrec_eq(f, g, bc=iasecross()):
    n = f.size
    y = MT.iacdil(f,g,bc,n)
    return y

def iainfrec(m, f, Bc=iasecross()):
    h,w = Bc.shape
    hc,wc = h/2,w/2
    B = Bc.copy()
    off = np.transpose(B.nonzero()) - np.array([hc,wc])
    i = off[:,0] * w + off[:,1]
    Nids = MT.iaNlut(f.shape,off)
    x,y = np.where(Nids==f.size)
    Nids[x,y] = x
    Nids_pos = Nids[:,i<0] #de acordo com a convenção em Vincent93
    Nids_neg = Nids[:,i>0] #de acordo com a convenção em Vincent93

    I = f.flatten()
    J = m.flatten()
    D = np.nonzero(J)[0]
    V = np.zeros(f.size,np.bool) #para controle de inserção na fila
    fila =[]

    for p in D:
        Jq = J[p]
        for q in Nids_pos[p]:
            Jq = max(Jq,J[q])
            if (J[q] < J[p]) and (J[q] < I[q]) and ~V[p]:
                fila.append(p)
                V[p]=True
            J[p] = min(Jq,I[p])

    for p in D[::-1]:
        Jq = J[p]
        for q in Nids_neg[p]:
            Jq = max(Jq,J[q])
            if (J[q] < J[p]) and (J[q] < I[q]) and ~V[p]:
                fila.append(p)
                V[p]=True
            J[p] = min(Jq,I[p])

    while fila:
        p = fila.pop(0)
        for q in Nids[p]:
            if J[q]<J[p] and I[q]!=J[q]:
                J[q] = min(J[p],I[q])
                fila.append(q)

    return J.reshape(f.shape)




# -*- encoding: utf-8 -*-
# Module iaimg2se

from numpy import *
from string import upper

def iaimg2se(fd, FLAT="FLAT", f=None):
    from iaisbinary import iaisbinary
    from iaseshow import iaseshow
    from ialimits import ialimits

    fd = (fd > 0)
    #assert iaisbinary(fd),'First parameter must be binary'
    FLAT = upper(FLAT)
    if FLAT == 'FLAT':
        return iaseshow(fd)
    else:
        B = choose(fd, ( ialimits(int32([0]))[0]*ones(fd.shape),f) )
    B = iaseshow(int32(B),'NON-FLAT')

    return B




# -*- encoding: utf-8 -*-
# Module iahmax

from numpy import *
from iasecross import iasecross

def iahmax(f, h=1, Bc=iasecross()):
    from iasubm import iasubm
    from iainfrec import iainfrec

    g = iasubm(f,h)
    y = iainfrec(g,f,Bc);
    return y




# -*- encoding: utf-8 -*-
# Module iagradm

from numpy import *

def iagradm(f, Bdil=None, Bero=None):
    from iasubm import iasubm
    from iadil import iadil
    from iaero import iaero
    from iasecross import iasecross
    if Bdil is None: Bdil = iasecross()
    if Bero is None: Bero = iasecross()

    y = iasubm( iadil(f,Bdil),iaero(f,Bero))
    return y




# -*- encoding: utf-8 -*-
# Module iaseshow

from numpy import *
from string import upper

def iaseshow(B, option="NORMAL"):
    from ia870 import iaisbinary, iaintersec, iagray, iabinary
    from ia870 import iasedil, iaero, iabshow

    option = upper(option)
    if option=='NON-FLAT':
        y = array([0],int32)
        if iaisbinary(B):
            B = iaintersec( iagray(B,'int32'),0)
    elif option=='NORMAL':
        if iaisbinary(B):    y=iabinary([1])
        else:
           y=array([0],int32)
    elif option=='EXPAND':
        assert iaisbinary(B), 'This option is only available with flat SE'
        y = iasedil( iabinary([1]),B)
        b1= iabinary(y>=0)
        b0= b1<0
        b0[shape(b0)[0]/2, shape(b0)[1]/2] = 1
        y = iabshow(b1,y,b0)
        return y
    else:
        print 'iaseshow: not a valid flag: NORMAL, EXPAND or NON-FLAT'

    y = iasedil(y,B)
    return y




# -*- encoding: utf-8 -*-
# Module iaunpad

from numpy import *
from iasecross import iasecross

def iaunpad(f, B=iasecross()):
    from iamat2set import iamat2set
    from iaseshow import iaseshow

    i,v=iamat2set( iaseshow(B));
    mni=minimum.reduce(i)
    mxi=maximum.reduce(i)
    g = f[-mni[0]:f.shape[0]-mxi[0], -mni[1]:f.shape[1]-mxi[1]]

    return g




# -*- encoding: utf-8 -*-
# Module iacwatershed

from ia870 import iasecross, iaisbinary, ialabel, iasubm, iaero, iabinary

def iacwatershed(f, g, Bc=iasecross(), option='LINES'):
    from ipdp import se2offset

    if iaisbinary(g):
        g = ialabel(g, Bc)
    offset = se2offset(Bc)
    w = ift_m(f, offset, g)
    if option == 'LINES':
        w = iasubm(w, iaero(w))
        w = iabinary(w)
    return w

def ift_m(im, offsets, M):

    from ipdp import wsImage

    # initialise variables
    ws = wsImage(im)
    N, im, lab, D = ws.begin(offsets)

    wsM = wsImage(M)
    NM, imM, labM, DM = wsM.begin(offsets)

    # make the set
    Mset = dict()
    for p in D:
        if imM[p] > 0:
            if Mset.has_key(imM[p]):
                Mset[imM[p]].append(p)
            else:
                Mset[imM[p]] = [p]

    ift_k(ws, im, Mset.values(), N, D, lab)

    return ws.end()

# constants
MASK = -2

def ift_k(ws, im, M, N, D, lab):

    import numpy as np
    from ipdp import wsHeapQueue

    # create the working images
    done = ws.makeWorkCopy(False)
    c1 = ws.makeWorkCopy(np.inf)
    par = ws.makeWorkCopy(MASK)

    lab[:] = MASK

    queue = wsHeapQueue()

    for m in xrange(len(M)):
        for p in M[m]:
            c1[p] = im[p]
            lab[p] = m+1
            par[p] = p
            queue.push(p, im[p])

    while not queue.empty():
        p = queue.pop()
        done[p] = True
        for q in N(p):
            if done[q]:
                continue

            c = max(c1[p], im[q])
            if c < c1[q]:
                if c1[q] < np.inf:
                    if queue.contains(q, c1[q]):
                        queue.remove(q, c1[q])
                c1[q] = c
                lab[q] = lab[p]
                par[q] = p
                queue.push(q, c1[q])




# -*- encoding: utf-8 -*-
# Module iainterot

from numpy import *
from string import upper

def iainterot(Iab, theta=45, DIRECTION="CLOCKWISE"):
    from iase2hmt import iase2hmt
    from iaserot import iaserot


    DIRECTION = upper(DIRECTION)
    A,Bc = Iab
    if DIRECTION != 'CLOCKWISE':
        theta = 360 - theta
    Irot = iase2hmt( iaserot(A, theta),
                    iaserot(Bc,theta))


    return Irot




# -*- encoding: utf-8 -*-
# Module iacmp

from numpy import *

def iacmp(f1, oper, f2, oper1=None, f3=None):
    from iaintersec import iaintersec
    from iabinary import iabinary


    if   oper == '==':    y = (f1==f2)
    elif oper == '~=':    y = (f1!=f2)
    elif oper == '<=':    y = (f1<=f2)
    elif oper == '>=':    y = (f1>=f2)
    elif oper == '>':     y = (f1> f2)
    elif oper == '<':     y = (f1< f2)
    else:
        assert 0, 'oper must be one of: ==, ~=, >, >=, <, <=, it was:'+oper
    if oper1 != None:
        if   oper1 == '==':     y = iaintersec(y, f2==f3)
        elif oper1 == '~=':     y = iaintersec(y, f2!=f3)
        elif oper1 == '<=':     y = iaintersec(y, f2<=f3)
        elif oper1 == '>=':     y = iaintersec(y, f2>=f3)
        elif oper1 == '>':      y = iaintersec(y, f2> f3)
        elif oper1 == '<':      y = iaintersec(y, f2< f3)
        else:
            assert 0, 'oper1 must be one of: ==, ~=, >, >=, <, <=, it was:'+oper1

    y = iabinary(y)


    return y




# -*- encoding: utf-8 -*-
# Module iadist

from numpy import *
from string import upper
from iasecross import iasecross

def iadist(f, Bc=iasecross(), METRIC=None):
    from iagray import iagray
    from iaintersec import iaintersec
    from iaisequal import iaisequal
    from iaero import iaero
    from iasebox import iasebox

    if METRIC is not None: METRIC = upper(METRIC)
    f = iagray(f,'uint16')
    y = iaintersec(f,0)
    if (METRIC == 'EUCLIDEAN') or (METRIC == 'EUC2'):
        f = int32(f)
        b = int32(zeros((3,3)))
        i=1
        while any(f != y):
            a4,a2 = -4*i+2, -2*i+1
            b = int32([[a4,a2,a4],
                       [a2, 0,a2],
                       [a4,a2,a4]])
            y=f
            i=i+1
            f = iaero(f,b)
        if METRIC == 'EUCLIDEAN':
            y = uint16(sqrt(f)+0.5)
    else:
        if iaisequal(Bc, iasecross()):
            b = int32([[-2147483647,  -1, -2147483647],
                       [         -1,   0,          -1],
                       [-2147483647,  -1, -2147483647]])
        elif iaisequal(Bc, iasebox()):
            b = int32([[-1,-1,-1],
                       [-1, 0,-1],
                       [-1,-1,-1]])
        else: b = Bc
        while any(f != y):
            y=f
            f = iaero(f,b)
    return y




# -*- encoding: utf-8 -*-
# Module iasetrans

from numpy import *

def iasetrans(Bi, t):
    from iamat2set import iamat2set
    from iaset2mat import iaset2mat


    x,v=iamat2set(Bi)
    Bo = iaset2mat((x+t,v))
    Bo = Bo.astype(Bi.dtype)

    return Bo




# -*- encoding: utf-8 -*-
# Module iagplot

from numpy import *

def iagplot(plotitems=[], commands=[], ptitle="", xlabel="", ylabel="", size=(320,240)):


    import tempfile, shutil
    import adpil
    #
    tmpdir = tempfile.mkdtemp('', 'iagplot_')
    if not os.path.isdir(tmpdir):
        raise IOError('Cannot create temporary directory')
    #
    imgfilename = 'iagplot_IMG.png'
    cmdfilename = 'iagplot_CMD.cmd'
    dataprefix  = 'iagplot_DATA'
    #
    currdir = os.getcwd()
    os.chdir(tmpdir)
    #
    try:
        cmdf = open(cmdfilename, 'w')
        ver = get_gnuver()
        if ver.startswith('4.'):
            print >> cmdf, 'set terminal png size %d,%d transparent small' % size
        else:
            print >> cmdf, 'set terminal png transparent small picsize %d %d' % size
        print >> cmdf, 'set output "%s"' % imgfilename.replace('\\', '/')
        print >> cmdf, 'set title "%s"'  % ptitle
        print >> cmdf, 'set xlabel "%s"' % xlabel
        print >> cmdf, 'set ylabel "%s"' % ylabel
        for cmd in commands:
            print >> cmdf, cmd
        NN = len(plotitems)
        strlist = []
        for ii in range(NN):
            item = plotitems[ii]
            datafile = '%s_%d.dat' % (dataprefix, ii+1)
            dataf = open(datafile, 'w')
            style = ''
            N = len(item)
            if N == 1:
                yy = asarray(item[0]).flat
                xx = arange(len(yy))
            else:
                yy = asarray(item[1]).flat
                if item[0] is None or len(item[0]) != len(yy):
                    xx = arange(len(yy))
                else:
                    xx = asarray(item[0]).flat
            if N == 3:
                style = 'title "%s" with lines' % (item[2],)
            elif N == 4:
                style = 'title "%s" with %s' % (item[2], item[3])
            else:
                style = 'title "" with lines'

            for i in range(len(yy)):
                print >> dataf, '%f %f' % (xx[i], yy[i])
            dataf.close();
            strlist.append('"%s" %s' % (datafile, style))

        print >> cmdf, 'plot', ', '.join(strlist)
        print >> cmdf
        cmdf.close()
        #
        os.system('"%s" %s' % (GNUPLOT, cmdfilename))
        img = adpil.adread(imgfilename)
        #
    finally:
        os.chdir(currdir)
        shutil.rmtree(tmpdir, True)

    return img




# -*- encoding: utf-8 -*-
# Module iainfgen

from numpy import *

def iainfgen(f, Iab):
    from iaunion import iaunion
    from iadil import iadil
    from ianeg import ianeg


    A, Bc = Iab
    y = iaunion( iadil(f, A), iadil( ianeg(f), Bc))

    return y




# -*- encoding: utf-8 -*-
# Module iadil

from numpy import *

def iadil(f, b=None):
    from iamat2set import iamat2set
    from ialimits import ialimits
    from iaisbinary import iaisbinary
    from iaintersec import iaintersec
    from iagray import iagray
    from iaadd4dil import iaadd4dil
    from iasecross import iasecross

    if b is None: b = iasecross()

    if len(f.shape) == 1: f = f[newaxis,:]
    h,w = f.shape
    x,v = iamat2set(b)
    if len(x)==0:
        y = (ones((h,w)) * ialimits(f)[0]).astype(f.dtype)
    else:
        if iaisbinary(v):
            v = iaintersec( iagray(v,'int32'),0)
        mh,mw = max(abs(x)[:,0]),max(abs(x)[:,1])
        y = (ones((h+2*mh,w+2*mw)) * ialimits(f)[0]).astype(f.dtype)
        for i in range(x.shape[0]):
            if v[i] > -2147483647:
                y[mh+x[i,0]:mh+x[i,0]+h, mw+x[i,1]:mw+x[i,1]+w] = maximum(
                    y[mh+x[i,0]:mh+x[i,0]+h, mw+x[i,1]:mw+x[i,1]+w], iaadd4dil(f,v[i]))
        y = y[mh:mh+h, mw:mw+w]

    return y




# -*- encoding: utf-8 -*-
# Module iacloserec

from numpy import *
from iasecross import iasecross

def iacloserec(f, bdil=iasecross(), bc=iasecross()):
    from iasuprec import iasuprec
    from iadil import iadil

    return iasuprec( iadil(f,bdil),f,bc)




# -*- encoding: utf-8 -*-
# Module iase2off

import numpy as np
def iase2off(Bc,option='neigh'):
    '''Converts structuring element to list of neighbor offsets in graph image'''
    h,w = Bc.shape
    hc,wc = h/2,w/2
    B = Bc.copy()
    B[hc,wc] = 0  # remove origin
    off = np.transpose(B.nonzero()) - np.array([hc,wc])
    if option == 'neigh':
        return off  # 2 columns x n. of neighbors rows
    elif option == 'fw':
        i = off[:,0] * w + off[:,1]
        return off[i>0,:]  # only neighbors higher than origin in raster order
    elif option == 'bw':
        i = off[:,0] * w + off[:,1]
        return off[i<0,:]  # only neighbors less than origin in raster order
    else:
        assert 0,'options are neigh, fw or bw. It was %s'% option
        return None




# -*- encoding: utf-8 -*-
# Module iaareaclose

from numpy import *
from iasecross import iasecross

def iaareaclose(f, a, Bc=iasecross()):
    from ianeg import ianeg
    from iaareaopen import iaareaopen

    y = ianeg( iaareaopen( ianeg(f),a,Bc))
    return y




# -*- encoding: utf-8 -*-
# Module MainPage

#
# Toolbox ia870
#

from iaaddm import iaaddm
from iaareaclose import iaareaclose
from iaareaopen import iaareaopen
from iaasf import iaasf
from iaasfrec import iaasfrec
from iabinary import iabinary
from iablob import iablob
from iabshow import iabshow
from iacbisector import iacbisector
from iacdil import iacdil
from iacenter import iacenter
from iacero import iacero
from iaclohole import iaclohole
from iaclose import iaclose
from iacloserec import iacloserec
from iacloserecth import iacloserecth
from iacloseth import iacloseth
from iacmp import iacmp
from iaconcat import iaconcat
from iacthick import iacthick
from iacthin import iacthin
from iadil import iadil
from iadist import iadist
from iadrawv import iadrawv
from iaedgeoff import iaedgeoff
from iaendpoints import iaendpoints
from iaero import iaero
from iaframe import iaframe
from iagdtshow import iagdtshow
from iagdist import iagdist
from iaglblshow import iaglblshow
from iagplot import iagplot
from iagradm import iagradm
from iagrain import iagrain
from iagray import iagray
from iagshow import iagshow
from iagsurf import iagsurf
from iahmax import iahmax
from iahmin import iahmin
from iahomothick import iahomothick
from iahomothin import iahomothin
from iaimg2se import iaimg2se
from iainfcanon import iainfcanon
from iainfgen import iainfgen
from iainfrec import iainfrec
from iainpos import iainpos
from iainterot import iainterot
from iaintersec import iaintersec
from iaintershow import iaintershow
from iais import iais
from iaisbinary import iaisbinary
from iaisequal import iaisequal
from iaislesseq import iaislesseq
from ialabel import ialabel
from ialabelflat import ialabelflat
from ialastero import ialastero
from ialimits import ialimits
from iamat2set import iamat2set
from iamaxleveltype import iamaxleveltype
from ianeg import ianeg
from iaNlut import iaNlut
from iaopen import iaopen
from iaopenrec import iaopenrec
from iaopenrecth import iaopenrecth
from iaopenth import iaopenth
from iaopentransf import iaopentransf
from iapad import iapad
from iaadd4dil import iaadd4dil
from iapad4n import iapad4n
from iapatspec import iapatspec
from iaregmax import iaregmax
from iaregmin import iaregmin
from iase2interval import iase2interval
from iase2hmt import iase2hmt
from iase2off import iase2off
from iasebox import iasebox
from iasecross import iasecross
from iasedisk import iasedisk
from iaseline import iaseline
from iaserot import iaserot
from iaset2mat import iaset2mat
from iaseshow import iaseshow
from iasesum import iasesum
from iasetrans import iasetrans
from iasereflect import iasereflect
from iaseintersec import iaseintersec
from iasedil import iasedil
from iaseunion import iaseunion
from iaskelm import iaskelm
from iaskelmrec import iaskelmrec
from iastats import iastats
from iasubm import iasubm
from iasupcanon import iasupcanon
from iasupgen import iasupgen
from iasuprec import iasuprec
from iasymdif import iasymdif
from iatext import iatext
from iathick import iathick
from iathin import iathin
from iathreshad import iathreshad
from iatoggle import iatoggle
from iaunpad import iaunpad
from iaunion import iaunion

# Toolbox watershed
from iawatershed import iawatershed
from iatz import iatz
from iacwatershed import iacwatershed




# -*- encoding: utf-8 -*-
# Module iatoggle

from numpy import *
from string import upper

def iatoggle(f, f1, f2, OPTION="GRAY"):
    from iabinary import iabinary
    from iasubm import iasubm
    from iagray import iagray
    from iaunion import iaunion
    from iaintersec import iaintersec
    from ianeg import ianeg


    y=iabinary( iasubm(f,f1),iasubm(f2,f))
    if upper(OPTION) == 'GRAY':
        t=iagray(y)
        y=iaunion( iaintersec( ianeg(t),f1),iaintersec(t,f2))


    return y




# -*- encoding: utf-8 -*-
# Module iagdist

from numpy import *
from iasecross import iasecross

def iagdist(f, g, Bc=iasecross(), METRIC=None):
    from ianeg import ianeg
    from iagray import iagray
    from iaintersec import iaintersec
    from iaisequal import iaisequal
    from iacero import iacero
    from iaaddm import iaaddm
    from iaunion import iaunion

    assert METRIC is None,'Does not support EUCLIDEAN'
    fneg,gneg = ianeg(f),ianeg(g)
    y = iagray(gneg,'uint16',1)
    ero = iaintersec(y,0)
    aux = y
    i = 1
    while (ero != aux).any():
        aux = ero
        ero = iacero(gneg,fneg,Bc,i)
        y   = iaaddm(y,iagray(ero,'uint16',1))
        i   = i + 1
    y = iaunion(y,iagray(ero,'uint16'))

    return y




# -*- encoding: utf-8 -*-
# Module iacero

from numpy import *
from iasecross import iasecross

def iacero(f, g, b=iasecross(), n=1):
    from iaunion import iaunion
    from iaero import iaero
    from iaisequal import iaisequal

    y = iaunion(f,g)
    for i in range(n):
        aux = y
        y = iaunion( iaero(y,b),g)
        if iaisequal(y,aux): break
    return y




# -*- encoding: utf-8 -*-
# Module iaopenrecth

from numpy import *

def iaopenrecth(f, bero=None, bc=None):
    from iasubm import iasubm
    from iaopenrec import iaopenrec
    from iasecross import iasecross
    if bero is None:
        bero = iasecross()
    if bc is None:
        bc = iasecross()

    y = iasubm(f, iaopenrec( f, bero, bc))

    return y




# -*- encoding: utf-8 -*-
# Module iahmin

from numpy import *
from iasecross import iasecross

def iahmin(f, h=1, Bc=iasecross()):
    from iaaddm import iaaddm
    from iasuprec import iasuprec

    g = iaaddm(f,h)
    y = iasuprec(g,f,Bc);
    return y




# -*- encoding: utf-8 -*-
# Module iaserot

from numpy import *
from string import upper

def iaserot(B, theta=45, DIRECTION="CLOCKWISE"):
    from iamat2set import iamat2set
    from iabinary import iabinary
    from iaset2mat import iaset2mat


    DIRECTION = upper(DIRECTION)
    if DIRECTION == "ANTI-CLOCKWISE":
       theta = -theta
    SA = iamat2set(B)
    theta = pi * theta/180
    (y,v)=SA
    if len(y)==0: return iabinary([0])
    x0 = y[:,1] * cos(theta) - y[:,0] * sin(theta)
    x1 = y[:,1] * sin(theta) + y[:,0] * cos(theta)
    x0 = int32((x0 +0.5)*(x0>=0) + (x0-0.5)*(x0<0))
    x1 = int32((x1 +0.5)*(x1>=0) + (x1-0.5)*(x1<0))
    x = transpose(array([transpose(x1),transpose(x0)]))
    BROT = iaset2mat((x,v))

    return BROT




# -*- encoding: utf-8 -*-
# Module iapad

from numpy import *
from iasecross import iasecross

def iapad(f, B=iasecross(), value=0):
    from iamat2set import iamat2set
    from iaseshow import iaseshow

    i,v=iamat2set( iaseshow(B));
    mni=i.min(axis=0)
    mxi=i.max(axis=0)
    f = asarray(f)
    if size(f.shape) == 1: f = f[newaxis,:]
    g = (value * ones(array(f.shape)+mxi-mni)).astype(f.dtype)
    g[-mni[0]:g.shape[0]-mxi[0], -mni[1]:g.shape[1]-mxi[1]] = f

    return g




# -*- encoding: utf-8 -*-
# Module iaframe

from numpy import *

def iaframe(f, WT=1, HT=1, DT=0, k1=None, k2=None):
    from iaunion import iaunion
    from iaintersec import iaintersec
    from ialimits import ialimits

    if k1 is None: k1 = ialimits(f)[1]
    if k2 is None: k2 = ialimits(f)[0]
    assert len(f.shape)==2,'Supports 2D only'
    y = iaintersec(f,k2)
    y[:,0:WT] = k1
    y[:,-WT:] = k1
    y[0:HT,:] = k1
    y[-HT:,:] = k1
    return y




# -*- encoding: utf-8 -*-
# Module iaglblshow

from numpy import *
from numpy.random import rand

def iaglblshow(X, border=0.0):
    from iastats import iastats
    from iaconcat import iaconcat

    iain = iastats(X,'min')
    iaax = iastats(X,'max')
    ncolors = iaax - iain + 1
    R = rand(ncolors)*255
    G = rand(ncolors)*255
    B = rand(ncolors)*255
    if iain == 0:
       R[0],G[0],B[0] = 0,0,0
    r=resize(take(R.astype(uint8), X.flat - iain),X.shape)
    g=resize(take(G.astype(uint8), X.flat - iain),X.shape)
    b=resize(take(B.astype(uint8), X.flat - iain),X.shape)
    Y=iaconcat('d',r,g,b)
    return Y




# -*- encoding: utf-8 -*-
# Module iaareaopen

import numpy as np
import ia870 as MT
from iasecross import iasecross
def iaareaopen_eq(f, a, Bc=iasecross()):

    if f.dtype == np.bool:
      fr = MT.ialabel(f,Bc)      # binary area open, use area measurement
      g = MT.iablob(fr,'area')
      y = g >= a
    else:
      y = np.zeros_like(f)
      k1 = f.min()
      k2 = f.max()
      for k in xrange(k1,k2+1):   # gray-scale, use thresholding decomposition
        fk = (f >= k)
        fo = MT.iaareaopen(fk,a,Bc)
        if not fo.any():
          break
        y = MT.iaunion(y, MT.iagray(fo,f.dtype,k))
    return y
def find_area(area, i):
    lista = []
    while area[i] >= 0:
        lista.append(i)
        i = area[i]
    area[lista] = i
    return i

def iaareaopen(f,a,Bc=iasecross()):
    a = -a
    s = f.shape
    g = np.zeros_like(f).ravel()
    f1 = np.concatenate((f.ravel(),np.array([0])))
    area = -np.ones((f1.size,), np.int32)
    N = MT.iaNlut(s, MT.iase2off(Bc))
    pontos = f1.nonzero()[0]
    pontos = pontos[np.lexsort((np.arange(0,-len(pontos),-1),f1[pontos]))[::-1]]
    for p in pontos:
        for v in N[p]:
            if f1[p] < f1[v] or (f1[p] == f1[v] and v < p):
                rv = find_area(area, v)
                if rv != p:
                    if area[rv] > a or f1[p] == f1[rv]:
                        area[p] = area[p] + area[rv]
                        area[rv] = p
                    else:
                        area[p] = a
    for p in pontos[::-1]:
        if area[p] >= 0:
            g[p] = g[area[p]]
        else:
            if area[p] <= a:
                g[p] = f1[p]
    return g.reshape(s)




# -*- encoding: utf-8 -*-
# Module ianeg

from numpy import *

def ianeg(f):
    from ialimits import ialimits

    if ialimits(f)[0] == (- ialimits(f)[1]):
       y = -f
    else:
       y = ialimits(f)[0] + ialimits(f)[1] - f
    y = y.astype(f.dtype)
    return y




# -*- encoding: utf-8 -*-
# Module iais

from numpy import *
from string import upper

def iais(f1, oper, f2=None, oper1=None, f3=None):
    from iaisbinary import iaisbinary
    from iaisequal import iaisequal
    from iaislesseq import iaislesseq
    from ianeg import ianeg
    from iathreshad import iathreshad
    from iabinary import iabinary


    if f2 == None:
        oper=upper(oper);
        if   oper == 'BINARY': return iaisbinary(f1)
        elif oper == 'GRAY'  : return not iaisbinary(f1)
        else:
            assert 0,'oper should be BINARY or GRAY, was'+oper
    elif oper == '==':    y = iaisequal(f1, f2)
    elif oper == '~=':    y = not iaisequal(f1,f2)
    elif oper == '<=':    y = iaislesseq(f1,f2)
    elif oper == '>=':    y = iaislesseq(f2,f1)
    elif oper == '>':     y = iaisequal( ianeg( iathreshad(f2,f1)),iabinary(1))
    elif oper == '<':     y = iaisequal( ianeg( iathreshad(f1,f2)),iabinary(1))
    else:
        assert 0,'oper must be one of: ==, ~=, >, >=, <, <=, it was:'+oper
    if oper1 != None:
        if   oper1 == '==': y = y and iaisequal(f2,f3)
        elif oper1 == '~=': y = y and (not iaisequal(f2,f3))
        elif oper1 == '<=': y = y and iaislesseq(f2,f3)
        elif oper1 == '>=': y = y and iaislesseq(f3,f2)
        elif oper1 == '>':  y = y and iaisequal( ianeg( iathreshad(f3,f2)),iabinary(1))
        elif oper1 == '<':  y = y and iaisequal( ianeg( iathreshad(f2,f3)),iabinary(1))
        else:
            assert 0,'oper1 must be one of: ==, ~=, >, >=, <, <=, it was:'+oper1


    return y




# -*- encoding: utf-8 -*-
# Module iaasf

from numpy import *
from string import upper

def iaasf(f, SEQ="OC", b=None, n=1):
    from iasesum import iasesum
    from iaopen import iaopen
    from iaclose import iaclose
    from iasecross import iasecross
    if b is None:
        b = iasecross(None)

    SEQ=upper(SEQ)
    y = f
    if SEQ == 'OC':
        for i in range(1,n+1):
            nb = iasesum(b,i)
            y = iaopen( iaclose(y,nb),nb)
    elif SEQ == 'CO':
        for i in range(1,n+1):
            nb = iasesum(b,i)
            y = iaclose( iaopen(y,nb),nb)
    elif SEQ == 'OCO':
        for i in range(1,n+1):
            nb = iasesum(b,i)
            y = iaopen( iaclose( iaopen(y,nb),nb),nb)
    elif SEQ == 'COC':
        for i in range(1,n+1):
            nb = iasesum(b,i)
            y = iaclose( iaopen( iaclose(y,nb),nb),nb)

    return y




# -*- encoding: utf-8 -*-
# Module iaaddm

from numpy import *

def iaaddm(f1, f2):
    from ialimits import ialimits

    if type(f2) is array:
        assert f1.dtype == f2.dtype, 'Cannot have different datatypes:'
    k1,k2 = ialimits(f1)
    y = clip(f1.astype(int64)+f2, k1, k2)
    y = y.astype(f1.dtype)
    return y




# -*- encoding: utf-8 -*-
# Module iagrain

import numpy as np
from string import upper

def iagrain(fr, f, measurement, option="image"):

    measurement = upper(measurement)
    option      = upper(option)
    if fr.ndim == 1: fr = fr[newaxis,:]
    n = fr.max()
    if option == 'DATA': y = np.empty((n,),np.float)
    else               : y = np.zeros(fr.shape)
    if measurement == 'MAX':
        for i in range(1,n+1):
            val = f[fr==i].max()
            if option == 'DATA': y[i-1] = val
            else               : y[fr==i] = val
    elif measurement == 'MIN':
        for i in range(1,n+1):
            val = f[fr==i].min()
            if option == 'DATA': y[i-1] = val
            else               : y[fr==i] = val
    elif measurement == 'SUM':
        for i in range(1,n+1):
            val = f[fr==i].sum()
            if option == 'DATA': y[i-1] = val
            else               : y[fr==i] = val
    elif measurement == 'MEAN':
        for i in range(1,n+1):
            val = f[fr==i].mean()
            if option == 'DATA': y[i-1] = val
            else               : y[fr==i] = val
    elif measurement == 'STD':
        for i in range(1,n+1):
            v = f[fr==i].std()
            if len(v) < 2: val = 0
            else         : val = v.std()
            if option == 'DATA': y[i-1] = val
            else               : y[fr==i] = val
    elif measurement == 'STD1':
        print "'STD1' is not implemented"
    else:
        print "Measurement should be 'MAX', 'MIN', 'MEAN', 'SUM', 'STD', 'STD1'."
    return y




# -*- encoding: utf-8 -*-
# Module iathin

from numpy import *
from string import upper

def iathin(f, Iab=None, n=-1, theta=45, DIRECTION="CLOCKWISE"):
    import ia870 as MT

    if Iab is None: Iab = MT.iahomothin()

    DIRECTION = upper(DIRECTION)
    assert MT.iaisbinary(f),'f must be binary image'
    if n == -1: n = product(f.shape)
    y = f
    zero = MT.iaintersec(f,0)
    for i in range(n):
        aux = zero
        for t in range(0,360,theta):
            sup = MT.iasupgen( y, MT.iainterot(Iab, t, DIRECTION))
            aux = MT.iaunion( aux, sup)
            #y -= sup
            y = MT.iasubm( y, sup)
        #if iaisequal(aux,zero): break
        if not aux.any(): break
    return y




# -*- encoding: utf-8 -*-
# Module iabshow

from numpy import *

def iabshow(f1, f2=None, f3=None, factor=17):
    from iabinary import iabinary
    from iaframe import iaframe
    from iadil import iadil
    from iaunion import iaunion
    from iasedisk import iasedisk
    from iaserot import iaserot
    from iasecross import iasecross
    from iasesum import iasesum

    assert f1.dtype == bool, 'f1 must be boolean image'
    factor = max(factor,9)
    hfactor = factor/2
    if size(f1.shape) == 1:
       f1 = f1[newaxis,:]
       if f2 != None: f2 = f2[newaxis,:]
       if f3 != None: f3 = f3[newaxis,:]
    bz = zeros(factor * array(f1.shape)).astype(bool)
    b0 = asarray(bz)
    b0[hfactor::factor,hfactor::factor] = f1
    fr1 = iaframe(zeros((factor,factor),bool))
    fr1 = iadil(b0,fr1)

    if f2 != None:
      assert f1.shape == f2.shape, 'f1 and f2 must have same shape'
      b1 = asarray(bz)
      b1[hfactor::factor,hfactor::factor] = f2
      fr2 = iadil(b1,iasedisk(hfactor - 4))
      fr1 = iaunion(fr1,fr2)

      if f3 != None:
        assert f1.shape == f3.shape, 'f1 and f3 must have same shape'
        bz[hfactor::factor,hfactor::factor] = f3
        fr3 = iadil(bz, iasesum(iaserot(iasecross(1),45),hfactor - 1))
        fr1 = iaunion(fr1,fr3)
    return fr1




# -*- encoding: utf-8 -*-
# Module iasupgen

from numpy import *

def iasupgen(f, INTER):
    from iaintersec import iaintersec
    from iaero import iaero
    from ianeg import ianeg


    A,Bc = INTER
    y = iaintersec( iaero( f, A),
                   iaero( ianeg(f), Bc))

    return y




# -*- encoding: utf-8 -*-
# Module iasebox

from numpy import *

def iasebox(r=1):
    from iasesum import iasesum
    from iabinary import iabinary


    B = iasesum( iabinary([[1,1,1],
                          [1,1,1],
                          [1,1,1]]),r)


    return B




# -*- encoding: utf-8 -*-
# Module iasecrop

def iasecrop(f):
    import numpy as np
    from ia870 import ialimits, iagray
    k1,k2 = ialimits(f)
    if not f.shape[0] & 1: f = np.vstack([k1 * np.ones((1,f.shape[1]),f.dtype),f])
    if not f.shape[1] & 1: f = np.hstack([k1 * np.ones((f.shape[0],1),f.dtype),f])
    h,w = f.shape
    h1 = (f != k1).max(1).nonzero()[0]
    dh = min([h1[0],h-h1[-1]-1])
    fh = f[dh:h-dh,:]
    w1 = (fh != k1).max(0).nonzero()[0]
    dw = min([w1[0],w-w1[-1]-1])
    return fh[:,dw:w-dw]




# -*- encoding: utf-8 -*-
# Module iacloserecth

from numpy import *
from iasecross import iasecross

def iacloserecth(f, bdil=iasecross(), bc=iasecross()):
    from iasubm import iasubm
    from iacloserec import iacloserec

    y = iasubm( iacloserec(f,bdil,bc), f)
    return y




# -*- encoding: utf-8 -*-
# Module iaero

from numpy import *

def iaero(f, b=None):
    from ianeg import ianeg
    from iadil import iadil
    from iasereflect import iasereflect
    from iasecross import iasecross

    if b is None: b = iasecross()
    y = ianeg( iadil( ianeg(f),iasereflect(b)))
    return y




# -*- encoding: utf-8 -*-
# Module iapatspec

from numpy import *

def iapatspec(f, type='OCTAGON', n=65535, Bc=None, Buser=None):
    from iaisbinary import iaisbinary
    from iaopentransf import iaopentransf
    from iahistogram import iahistogram
    from iasecross import iasecross
    if Bc is None:
        Bc = iasecross(None)
    if Buser is None:
        Buser = iasecross(None)

    assert iaisbinary(f),'Error: input image is not binary'
    g=iaopentransf(f,type,n,Bc,Buser)
    h=iahistogram(g)
    h=h[1:]
    return h




# -*- encoding: utf-8 -*-
# Module iacloseth

from numpy import *

def iacloseth(f, b=None):
    from iasubm import iasubm
    from iaclose import iaclose
    from iasecross import iasecross
    if b is None:
        b = iasecross()

    y = iasubm( iaclose(f,b), f)
    return y




# -*- encoding: utf-8 -*-
# Module iathick

from numpy import *
from string import upper

def iathick(f, Iab=None, n=-1, theta=45, DIRECTION="CLOCKWISE"):
    from iaisbinary import iaisbinary
    from iaintersec import iaintersec
    from iasupgen import iasupgen
    from iainterot import iainterot
    from iaunion import iaunion
    from iaisequal import iaisequal
    from iahomothick import iahomothick
    if Iab is None:
        Iab = iahomothick()

    DIRECTION = upper(DIRECTION)
    assert iaisbinary(f),'f must be binary image'
    if n == -1: n = product(f.shape)
    y = f
    zero = iaintersec(f,0)
    for i in range(n):
        aux = zero
        for t in range(0,360,theta):
            sup = iasupgen( y, iainterot(Iab, t, DIRECTION))
            aux = iaunion( aux, sup)
            y = iaunion( y, sup)
        if iaisequal(aux,zero): break

    return y




# -*- encoding: utf-8 -*-
# Module iaendpoints

from numpy import *
from string import upper

def iaendpoints(OPTION="LOOP"):
    from iase2hmt import iase2hmt
    from iabinary import iabinary


    Iab = None
    OPTION = upper(OPTION)
    if OPTION == 'LOOP':
        Iab = iase2hmt( iabinary([[0,0,0],
                                  [0,1,0],
                                  [0,0,0]]),
                        iabinary([[0,0,0],
                                  [1,0,1],
                                  [1,1,1]]))
    elif OPTION == 'HOMOTOPIC':
        Iab = iase2hmt( iabinary([[0,1,0],
                                  [0,1,0],
                                  [0,0,0]]),
                        iabinary([[0,0,0],
                                  [1,0,1],
                                  [1,1,1]]))


    return Iab




# -*- encoding: utf-8 -*-
# Module iacenter

from numpy import *

def iacenter(f, b=None):
    from iaasf import iaasf
    from iaunion import iaunion
    from iaintersec import iaintersec
    from iaisequal import iaisequal
    from iasecross import iasecross
    if b is None:
        b = iasecross(None)

    y = f
    diff = 0
    while not diff:
        aux = y
        beta1 = iaasf(y,'COC',b,1)
        beta2 = iaasf(y,'OCO',b,1)
        y = iaunion( iaintersec(y,beta1),beta2)
        diff = iaisequal(aux,y)

    return y




# -*- encoding: utf-8 -*-
# Module ialabel

import numpy as np
import ia870 as MT
from iasecross import iasecross

def ialabel(f, Bc=iasecross()):
    return ialabel_unionfind(f,Bc)

# Implementation using morphological reconstruction iainfrec
def ialabel_rec(f, Bc=iasecross()):
    assert MT.iaisbinary(f),'Can only label binary image'
    faux=f.copy()
    label = 1
    y = MT.iagray( f,'uint16',0)          # zero image (output)
    x = faux.ravel().nonzero()            # get list of unlabeled pixel
    while len(x[0]):
        fmark = np.zeros_like(f)
        fmark.flat[x[0][0]] = 1           # get the first unlabeled pixel
        r = MT.iainfrec( fmark, faux, Bc) # detects all pixels connected to it
        faux -= r                         # remove them from faux
        r = MT.iagray( r,'uint16',label)  # label them with the value label
        y = MT.iaunion( y, r)             # merge them with the labeled image
        label = label + 1
        x = faux.ravel().nonzero()        # get list of unlabeled pixel
    return y

# implementation by union find
def Find(x,parents): # uses path compression
   if parents[x] == x:
      return x
   else:
      parents[x] = Find(parents[x],parents)
      return parents[x]

def Union(n, p,parents):
   r = Find(n,parents)
   if r != p:
      parents[r] = p

def ialabel_unionfind(img,Bc):
   imshape = img.shape
   imsize  = img.size

   # Offsets and neighbors
   offsets = MT.iase2off(Bc,'fw')
   neighbors = MT.iaNlut(imshape,offsets)

   parents = np.arange(imsize,dtype = int)

   # Raster image and no-nzero pixels
   img = img.ravel()
   nonzero_nodes = np.nonzero(img)[0]
   img =  np.concatenate((img,np.array([0])))
   O = np.zeros(imsize, dtype = int)
   cur_label = 0
   # pass 1: backward scan, forward neighbors
   for p in nonzero_nodes[::-1]:
      for nb in neighbors[p]:
         if img[nb]:
            Union(nb,p,parents)
   # pass 2_1: root labeled
   pbool = parents[nonzero_nodes] == nonzero_nodes
   labels = np.arange(1,pbool.sum()+1)
   O[nonzero_nodes[pbool]] = labels
   # pass 2_2: labeling root descendants
   for p in nonzero_nodes[~pbool]:
       O[p] = O[parents[p]]
   return O.reshape(imshape)




# -*- encoding: utf-8 -*-
# Module iainfcanon

from numpy import *
from string import upper

def iainfcanon(f, Iab, theta=45, DIRECTION="CLOCKWISE"):
    from iaunion import iaunion
    from iainterot import iainterot
    from iaintersec import iaintersec
    from iainfgen import iainfgen

    DIRECTION = upper(DIRECTION)
    y = iaunion(f,1)
    for t in range(0,360,theta):
        Irot = iainterot( Iab, t, DIRECTION )
        y = iaintersec( y, iainfgen(f, Irot))

    return y




# -*- encoding: utf-8 -*-
# Module iacdil

from numpy import *
from iasecross import iasecross

def iacdil(f, g, b=iasecross(), n=1):
    from iaintersec import iaintersec
    from iadil import iadil
    from iaisequal import iaisequal

    y = iaintersec(f,g)
    for i in xrange(n):
        aux = y
        y = iaintersec( iadil(y,b),g)
        if iaisequal(y,aux): break
    return y




# -*- encoding: utf-8 -*-
# Module iaskelm

from numpy import *
from iasecross import iasecross
from string import upper

def iaskelm(f, B=iasecross(), option="binary"):
    from iaisbinary import iaisbinary
    from ialimits import ialimits
    from iagray import iagray
    from iaintersec import iaintersec
    from iasesum import iasesum
    from iaero import iaero
    from iaisequal import iaisequal
    from iaopenth import iaopenth
    from iasedil import iasedil
    from iaunion import iaunion
    from iabinary import iabinary
    from iapad import iapad
    from iaunpad import iaunpad

    assert iaisbinary(f),'Input binary image only'
    option = upper(option)
    f = iapad(f,B)
    print f
    k1,k2 = ialimits(f)
    y = iagray( iaintersec(f, k1),'uint16')
    iszero = asarray(y)
    nb = iasesum(B,0)
    for r in range(1,65535):
        ero = iaero( f, nb)
        if iaisequal(ero, iszero): break
        f1 = iaopenth( ero, B)
        nb = iasedil(nb, B)
        y = iaunion(y, iagray(f1,'uint16',r))
    if option == 'BINARY':
        y = iabinary(y)
    y = iaunpad(y,B)
    return y




# -*- encoding: utf-8 -*-
# Module iahomothin

from numpy import *

def iahomothin():
    from iase2hmt import iase2hmt
    from iabinary import iabinary


    Iab = iase2hmt( iabinary([[0,0,0],
                             [0,1,0],
                             [1,1,1]]),
                   iabinary([[1,1,1],
                             [0,0,0],
                             [0,0,0]]))


    return Iab




# -*- encoding: utf-8 -*-
# Module iagsurf

def iagsurf(f, vx=1, vy=0, vz=1):

    import numpy as np
    import ia636
    zd = ia636.iapad(f).astype(float)
    gv = zd[1:-1,2:] - zd[1:-1,1:-1]
    gv /= gv.max()
    #adshow(ia636.ianormalize(gv))
    gh = zd[2:,1:-1] - zd[1:-1,1:-1]
    gh /= gh.max()
    #adshow(ia636.ianormalize(gh))
    gz = 1.0/(gv**2 + gh**2 + 1)
    #adshow(ia636.ianormalize(gz))
    gv *= gz
    gh *= gz
    #adshow(ia636.ianormalize(gv))
    #adshow(ia636.ianormalize(gh))
    v = np.sqrt(vx*vx + vy*vy + vz*vz)
    vx = vx/v; vy = vy/v; vz = vz/v;
    gv *= vx
    gh *= vy
    gz *= vz
    gz += gh + gv
    return ia636.ianormalize(gz)




# -*- encoding: utf-8 -*-
# Module iapad4n

from numpy import *

def iapad4n(f, Bc, value, scale=1):
    from iaseshow import iaseshow

    if type(Bc) is not array:
      Bc = iaseshow(Bc)
    Bh, Bw = Bc.shape
    assert Bh%2 and Bw%2, 'structuring element must be odd sized'
    ch, cw = scale * Bh/2, scale * Bw/2
    g = value * ones( f.shape + scale * (array(Bc.shape) - 1))
    g[ ch: -ch, cw: -cw] = f
    y = g.astype( f.dtype.char)


    return y




# -*- encoding: utf-8 -*-
# Module iamat2set

from numpy import *

def iamat2set(A):
    from ialimits import ialimits


    if len(A.shape) == 1: A = A[newaxis,:]
    offsets = nonzero(ravel(A) - ialimits(A)[0])
    if type(offsets) == type(()):
        offsets = offsets[0]        # for compatibility with numarray
    if len(offsets) == 0: return ([],[])
    (h,w) = A.shape
    x = range(2)
    x[0] = offsets/w - (h-1)/2
    x[1] = offsets%w - (w-1)/2
    x = transpose(x)
    CV = x,ravel(A)[offsets]
    return CV




# -*- encoding: utf-8 -*-
# Module iaskelmrec

from numpy import *

def iaskelmrec(f, B=None):
    from iabinary import iabinary
    from iaintersec import iaintersec
    from iadil import iadil
    from iaunion import iaunion
    from iasecross import iasecross
    if B is None:
        B = iasecross(None)

    y = iabinary( iaintersec(f, 0))
    for r in range(max(ravel(f)),1,-1):
        y = iadil( iaunion(y,iabinary(f,r)), B)
    y = iaunion(y, iabinary(f,1))
    return y




# -*- encoding: utf-8 -*-
# Module iaclohole

from numpy import *
from iasecross import iasecross

def iaclohole(f, Bc=iasecross()):
    from iaframe import iaframe
    from ianeg import ianeg
    from iainfrec import iainfrec

    delta_f = iaframe(f)
    return ianeg( iainfrec( delta_f, ianeg(f), Bc))




# -*- encoding: utf-8 -*-
# Module iasymdif

from numpy import *

def iasymdif(f1, f2):
    from iaunion import iaunion
    from iasubm import iasubm
    y = iaunion( iasubm(f1,f2),iasubm(f2,f1))
    return y




# -*- encoding: utf-8 -*-
# Module iatz

from ia870 import iasecross

# constants
MASK = -2
TIE_ZONE = 0

def iatz(f, Bc=iasecross()):
    from ipdp import se2offset

    offset = se2offset(Bc)
    return tieZone(f, offset)



def tieZone(im, offsets):

    import numpy as np
    from ipdp import wsImage
    from ipdp import findMinima
    from ipdp import wsHeapQueue

    # initialise variables
    ws = wsImage(im)
    N, im, lab, D = ws.begin(offsets)

    # find minima
    M = findMinima(im, N, D)

    # create the working images
    done = ws.makeWorkCopy(False)
    c1 = ws.makeWorkCopy(np.inf)
    c2 = ws.makeWorkCopy(0)
    par = ws.makeWorkCopy(MASK)
    lab[:] = MASK

    queue = wsHeapQueue()

    for m in xrange(len(M)):
        for p in M[m]:
            c1[p] = im[p]
            lab[p] = m+1
            par[p] = p
            queue.push(p, im[p])

    while not queue.empty():
        p = queue.pop()
        done[p] = True
        for q in N(p):
            if done[q]:
                continue

            c = max(c1[p], im[q])
            if c < c1[q]:
                if c1[q] < np.inf:
                    if queue.contains(q, c1[q]):
                        queue.remove(q, c1[q])
                c1[q] = c
                lab[q] = lab[p]
                par[q] = p
                queue.push(q, c1[q])
                if c == c1[p]:
                    c2[q] = c2[p] + 1
            elif c == c1[q] and lab[q] != lab[p]:
                if c == c1[p]:
                    if c2[q] == c2[p] + 1:
                        lab[q] = TIE_ZONE
                else:
                    lab[q] = TIE_ZONE

    return ws.end()




# -*- encoding: utf-8 -*-
# Module iasedil

from numpy import *

def iasedil(B1, B2):
    from ia870 import ialimits, iagray, iabinary, iaisbinary, iamat2set, iaadd4dil, iasetrans, iaseunion

    assert (iaisbinary(B1) or (B1.dtype == int32) or (B1.dtype == int64) or (B1.dtype == float64)) and \
           (iaisbinary(B2) or (B2.dtype == int32) or (B2.dtype == int64) or (B2.dtype == float64)), \
           'iasedil: s.e. must be binary, int32, int64 or float64'
    if len(B1.shape) == 1: B1 = B1[newaxis,:]
    if len(B2.shape) == 1: B2 = B2[newaxis,:]
    if B1.dtype=='bool' and B2.dtype == 'bool':
       Bo = iabinary([0])
    else:
       Bo = array(ialimits(B1)[0]).reshape(1)
       if iaisbinary(B1):
          Bo = array(ialimits(B2)[0]).reshape(1)
          B1 = iagray(B1,B2.dtype,0)
       if iaisbinary(B2):
          Bo = array(ialimits(B1)[0]).reshape(1)
          B2 = iagray(B2,B1.dtype,0)
    x,v = iamat2set(B2)
    if len(x):
        for i in range(x.shape[0]):
            s = iaadd4dil(B1,v[i])
            st= iasetrans(s,x[i])
            Bo = iaseunion(Bo,st)
    return Bo




# -*- encoding: utf-8 -*-
# Module iasedisk

from numpy import *
from string import upper

def iasedisk(r=3, DIM="2D", METRIC="EUCLIDEAN", FLAT="FLAT", h=0):
    from iabinary import iabinary
    from iasecross import iasecross
    from iasedil import iasedil
    from iasesum import iasesum
    from iasebox import iasebox
    from iaintersec import iaintersec
    from iagray import iagray

    METRIC = upper(METRIC)
    FLAT   = upper(FLAT)
    assert DIM=='2D','Supports only 2D structuring elements'
    if FLAT=='FLAT': y = iabinary([1])
    else:            y = int32([h])
    if r==0: return y
    if METRIC == 'CITY-BLOCK':
        if FLAT == 'FLAT':
            b = iasecross(1)
        else:
            b = int32([[-2147483647, 0,-2147483647],
                       [          0, 1,          0],
                       [-2147483647, 0,-2147483647]])
        return iasedil(y,iasesum(b,r))
    elif METRIC == 'CHESSBOARD':
        if FLAT == 'FLAT':
            b = iasebox(1)
        else:
            b = int32([[1,1,1],
                       [1,1,1],
                       [1,1,1]])
        return iasedil(y,iasesum(b,r))
    elif METRIC == 'OCTAGON':
        if FLAT == 'FLAT':
            b1,b2 = iasebox(1),iasecross(1)
        else:
            b1 = int32([[1,1,1],[1,1,1],[1,1,1]])
            b2 = int32([[-2147483647, 0,-2147483647],
                        [          0, 1,          0],
                        [-2147483647, 0,-2147483647]])
        if r==1: return b1
        else:    return iasedil( iasedil(y,iasesum(b1,r/2)) ,iasesum(b2,(r+1)/2))
    elif METRIC == 'EUCLIDEAN':
        v = arange(-r,r+1)
        x = resize(v, (len(v), len(v)))
        y = transpose(x)
        Be = iabinary(sqrt(x*x + y*y) <= (r+0.5))
        if FLAT=='FLAT':
            return Be
        be = h + int32( sqrt( maximum((r+0.5)*(r+0.5) - (x*x) - (y*y),0)))
        be = iaintersec( iagray(Be,'int32'),be)
        return be
    else:
        assert 0,'Non valid metric'


    return B




# -*- encoding: utf-8 -*-
# Module iaregmax

from numpy import *
from iasecross import iasecross

def iaregmax(f, Bc=iasecross()):
    from iasubm import iasubm
    from iahmax import iahmax
    from iabinary import iabinary
    from iaregmin import iaregmin
    from ianeg import ianeg

    y = iasubm(f, iahmax(f,1,Bc))
    return iabinary(y)
    #return iaregmin( ianeg(f),Bc)




# -*- encoding: utf-8 -*-
# Module iamaxleveltype

from numpy import *

def iamaxleveltype(TYPE='uint8'):

    max = 0
    if   TYPE == 'uint8'  : max=255
    elif TYPE == 'binary' : max=1
    elif TYPE == 'uint16' : max=65535
    elif TYPE == 'int32'  : max=2147483647
    elif TYPE == 'float64'  : max=inf
    else:
        assert 0, 'does not support this data type:'+TYPE

    return max




# -*- encoding: utf-8 -*-
# Module ialastero

from numpy import *
from iasecross import iasecross

def ialastero(f, B=iasecross()):
    from iaisbinary import iaisbinary
    from iadist import iadist
    from iaregmax import iaregmax

    assert iaisbinary(f),'Can only process binary images'
    dt = iadist(f,B)
    y = iaregmax(dt,B)
    return y




# -*- encoding: utf-8 -*-
# Module iaopenth

from numpy import *

def iaopenth(f, b=None):
    from iasubm import iasubm
    from iaopen import iaopen
    from iasecross import iasecross
    if b is None:
        b = iasecross()

    y = iasubm(f, iaopen(f,b))

    return y




# -*- encoding: utf-8 -*-
# Module iaopen

from numpy import *

def iaopen(f, b=None):
    from iadil import iadil
    from iaero import iaero
    from iasecross import iasecross
    if b is None:
        b = iasecross()

    y = iadil( iaero(f,b),b)

    return y




# -*- encoding: utf-8 -*-
# Module iaclose

from numpy import *

def iaclose(f, b=None):
    from iaero import iaero
    from iadil import iadil
    from iasecross import iasecross
    if b is None:
        b = iasecross()

    y = iaero( iadil(f,b),b)

    return y




# -*- encoding: utf-8 -*-
# Module iathreshad

from numpy import *

def iathreshad(f, f1, f2=None):
    from iabinary import iabinary

    if f2 is None:
      y = (f1 <= f)
    else:
      y = ((f1 <= f) & (f <= f2))

    return y




# -*- encoding: utf-8 -*-
# Module iacbisector

from numpy import *

def iacbisector(f, B, n):
    from iaintersec import iaintersec
    from iasesum import iasesum
    from iaero import iaero
    from iacdil import iacdil
    from iasubm import iasubm
    from iaunion import iaunion

    y = iaintersec(f,0)
    for i in range(n):
        nb = iasesum(B,i)
        nbp = iasesum(B,i+1)
        f1 = iaero(f,nbp)
        f2 = iacdil(f1,f,B,n)
        f3 = iasubm( iaero(f,nb),f2)
        y  = iaunion(y,f3)
    return y




# -*- encoding: utf-8 -*-
# Module iawatershed

from ia870 import iasecross, iasubm, iaero, iabinary

def iawatershed(f, Bc=iasecross(), option='LINES'):
    from ipdp import se2offset

    offset = se2offset(Bc)
    w = connectedComponents(f, offset)
    if option == 'LINES':
        w = iasubm(w, iaero(w))
        w = iabinary(w)
    return w

# constants
MASK = -2
PLATEAU = -1

def connectedComponents(im, offsets):

    from ipdp import wsImage
    from ipdp import wsQueue

    # initialise variables
    ws = wsImage(im)
    N, im, lab, D = ws.begin(offsets)

    lab[:] = MASK
    adr = ws.makeWorkCopy(0)

    queue = wsQueue()

    def find(p):
        q = p
        while adr[q] != q:
            q = adr[q]
        u = p
        while adr[u] != q:
            v = adr[u]
            adr[u] = q
            u = v
        return q

    # step 1
    for p in D:

        q = p
        for u in N(p):

            if im[u] < im[q]:
                q = u

        if q != p:
            adr[p] = q
        else:
            adr[p] = PLATEAU

    # step 2
    for p in D:
        if adr[p] != PLATEAU:
            continue

        for q in N(p):
            if adr[q] == PLATEAU or im[q] != im[p]:
                continue

            queue.push(q)

    while not queue.empty():
        p = queue.pop()
        for q in N(p):
            if adr[q] != PLATEAU or im[q] != im[p]:
                continue

            adr[q] = p
            queue.push(q)

    # step 3
    for p in D:
        if adr[p] != PLATEAU:
            continue

        adr[p] = p

        for q in N(p):
            if q > p or im[q] != im[p]:
                continue

            u = find(p)
            v = find(q)
            adr[u] = adr[v] = min(u,v)

    # step 4
    basins = 1
    for p in D:

        r = find(p)
        adr[p] = r
        if lab[r] == MASK:
            lab[r] = basins
            basins += 1
        lab[p] = lab[r]

    return ws.end()




# -*- encoding: utf-8 -*-
# Module iacthin

from numpy import *
from string import upper

def iacthin(f, g, Iab=None, n=-1, theta=45, DIRECTION="CLOCKWISE"):
    from iaisbinary import iaisbinary
    from iasupgen import iasupgen
    from iainterot import iainterot
    from iaunion import iaunion
    from iasubm import iasubm
    from iaisequal import iaisequal
    from iahomothin import iahomothin
    if Iab is None:
        Iab = iahomothin()

    DIRECTION = upper(DIRECTION)
    assert iaisbinary(f),'f must be binary image'
    if n == -1: n = product(f.shape)
    y = f
    old = y
    for i in range(n):
        for t in range(0,360,theta):
            sup = iasupgen( y, iainterot(Iab, t, DIRECTION))
            y = iaunion( iasubm( y, sup),g)
        if iaisequal(old,y): break
        old = y

    return y




# -*- encoding: utf-8 -*-
# Module iahomothick

from numpy import *

def iahomothick():
    from iase2hmt import iase2hmt
    from iabinary import iabinary


    Iab = iase2hmt( iabinary([[1,1,1],
                             [0,0,0],
                             [0,0,0]]),
                   iabinary([[0,0,0],
                             [0,1,0],
                             [1,1,1]]))


    return Iab




# -*- encoding: utf-8 -*-
# Module iasuprec

from numpy import *

def iasuprec(f, g, Bc=None):
    from iacero import iacero
    from iasecross import iasecross
    if Bc is None:
        Bc = iasecross(None)


    n = product(f.shape)
    y = iacero(f,g,Bc,n);

    return y




# -*- encoding: utf-8 -*-
# Module iacthick

from numpy import *
from string import upper

def iacthick(f, g, Iab=None, n=-1, theta=45, DIRECTION="CLOCKWISE"):
    from iaisbinary import iaisbinary
    from iasupgen import iasupgen
    from iainterot import iainterot
    from iaintersec import iaintersec
    from iaunion import iaunion
    from iaisequal import iaisequal
    from iahomothick import iahomothick
    if Iab is None:
        Iab = iahomothick()

    DIRECTION = upper(DIRECTION)
    assert iaisbinary(f),'f must be binary image'
    if n == -1: n = product(f.shape)
    y = f
    old = y
    for i in range(n):
        for t in range(0,360,theta):
            sup = iasupgen( y, iainterot(Iab, t, DIRECTION))
            y = iaintersec( iaunion( y, sup),g)
        if iaisequal(old,y): break
        old = y

    return y




# -*- encoding: utf-8 -*-
# Module ialabelflat

import numpy as np
import ia870 as MT
from iasecross import iasecross

def ialabelflat(f, Bc=iasecross(), delta=0):
    fr = f.astype(np.int) + 1
    return ialabelflat_unionfind(fr,Bc,delta)

# implementation by union find
def Find(x,parents): # uses path compression
   if parents[x] == x:
      return x
   else:
      parents[x] = Find(parents[x],parents)
      return parents[x]

def Union(n, p,parents):
   r = Find(n,parents)
   if r != p:
      parents[r] = p

def ialabelflat_unionfind(img,Bc,delta):
   imshape = img.shape
   imsize  = img.size

   # Offsets and neighbors
   offsets = MT.iase2off(Bc,'fw')
   neighbors = MT.iaNlut(imshape,offsets)

   parents = np.arange(imsize,dtype = int)

   # Raster image and no-nzero pixels
   img = img.ravel()
   nonzero_nodes = np.nonzero(img)[0]
   img =  np.concatenate((img,np.array([0])))
   g = np.zeros(imsize, dtype = int)
   cur_label = 0
   # pass 1: backward scan, forward neighbors
   for p in nonzero_nodes[::-1]:
      v = img[p]
      for nb in neighbors[p]:
         if img[nb] and (abs(v - img[nb]) <= delta):
            Union(nb,p,parents)
   # pass 2_1: root labeled
   pbool = parents[nonzero_nodes] == nonzero_nodes
   labels = np.arange(1,pbool.sum()+1)
   g[nonzero_nodes[pbool]] = labels
   # pass 2_2: labeling root descendants
   for p in nonzero_nodes[~pbool]:
       g[p] = g[parents[p]]
   return g.reshape(imshape)




# -*- encoding: utf-8 -*-
# Module iagray

from numpy import *

def iagray(f, TYPE="uint8", k1=None):
    from ia870 import iabinary, iais, iamaxleveltype, ianeg, ialimits

    #f = (f > 0)
    ff = array([0],TYPE)
    kk1,kk2 = ialimits(ff)
    if k1!=None:
        kk2=k1
    if   TYPE == 'uint8'  : y = where(f,kk2,kk1).astype(uint8)
    elif TYPE == 'uint16' : y = where(f,kk2,kk1).astype(uint16)
    elif TYPE == 'int32'  : y = where(f,kk2,kk1).astype(int32)
    elif TYPE == 'int64'  : y = where(f,kk2,kk1).astype(int64)
    elif TYPE == 'float64': y = where(f,kk2,kk1).astype(float64)
    else:
        assert 0, 'type not supported:'+TYPE
    return y




# -*- encoding: utf-8 -*-
# Module iaadd4dil

from numpy import *

def iaadd4dil(f, c):
    from ialimits import ialimits

    if not c:
       return f
    if f.dtype == 'float64':
        y = f + c
    else:
        y = asarray(f,int64) + c
        k1,k2 = ialimits(f)
        y = ((f==k1) * k1) + ((f!=k1) * y)
        y = clip(y,k1,k2)
    a = y.astype(f.dtype)
    return a




# -*- encoding: utf-8 -*-
# Module iaopentransf

from numpy import *
from string import find, upper
from iasecross import iasecross

def iaopentransf(f, type='OCTAGON', n=65535, Bc=iasecross(), Buser=iasecross()):
    from iaisbinary import iaisbinary
    from iabinary import iabinary
    from iaisequal import iaisequal
    from iaopen import iaopen
    from iasesum import iasesum
    from iasedisk import iasedisk
    from iaaddm import iaaddm
    from iagray import iagray
    from iagrain import iagrain
    from ialabel import ialabel

    assert iaisbinary(f),'Error: input image is not binary'
    type = upper(type)
    rec_flag = find(type,'-REC')
    if rec_flag != -1:
        type = type[:rec_flag] # remove the -rec suffix
    flag = not ((type == 'OCTAGON')  or
                (type == 'CHESSBOARD') or
                (type == 'CITY-BLOCK'))
    if not flag:
        n  = min(n,min(f.shape))
    elif  type == 'LINEAR-H':
        se = iabinary([1, 1, 1])
        n  = min(n,f.shape[1])
    elif  type =='LINEAR-V':
        se = iabinary([[1],[1],[1]])
        n  = min(n,f.shape[0])
    elif  type == 'LINEAR-45R':
        se = iabinary([[0, 0, 1],[0, 1, 0],[1, 0, 0]])
        n  = min(n,min(f.shape))
    elif  type == 'LINEAR-45L':
        se = iabinary([[1, 0, 0],[0, 1, 0],[0, 0, 1]])
        n  = min(n,min(f.shape))
    elif  type == 'USER':
        se = Buser
        n  = min(n,min(f.shape))
    else:
        print 'Error: only accepts OCTAGON, CHESSBOARD, CITY-BLOCK, LINEAR-H, LINEAR-V, LINEAR-45R, LINEAR-45L, or USER as type, or with suffix -REC.'
        return []
    k = 0
    y = uint16(zeros(f.shape))
    a = iabinary([1])
    z = iabinary([0])
    while not ( iaisequal(a,z) or (k>=n)):
        print 'processing r=',k
        if flag:
            a = iaopen(f,iasesum(se,k))
        else:
            a = iaopen(f,iasedisk(k,'2D',type))
        y = iaaddm(y, iagray(a,'uint16',1))
        k = k+1
    if rec_flag != -1:
        y = iagrain( ialabel(f,Bc),y,'max')

    return y




# -*- encoding: utf-8 -*-
# Module iasecross

from numpy import *

def iasecross(r=1):
    from iasesum import iasesum
    from iabinary import iabinary

    B = iasesum( iabinary([[0,1,0],
                           [1,1,1],
                           [0,1,0]]),r)
    return B




# -*- encoding: utf-8 -*-
# Module iasesum

import numpy as np

def iasesum(B, N=1):
    from ia870 import iaisbinary, iabinary, iasedil

    if N==0:
        if iaisbinary(B): return iabinary([1])
        else:             return array([0],int32) # identity
    NB = B
    for i in range(N-1):
        NB = iasedil(NB,B)
    return NB




# -*- encoding: utf-8 -*-
# Module iaopenrec

from numpy import *
from iasecross import iasecross

def iaopenrec(f, bero=iasecross(), bc=iasecross()):
    from iainfrec import iainfrec
    from iaero import iaero

    return iainfrec( iaero(f,bero),f,bc)




# -*- encoding: utf-8 -*-
# Module iasupcanon

from numpy import *
from string import upper

def iasupcanon(f, Iab, theta=45, DIRECTION="CLOCKWISE"):
    from iaintersec import iaintersec
    from iainterot import iainterot
    from iaunion import iaunion
    from iasupgen import iasupgen

    DIRECTION = upper(DIRECTION)
    y = iaintersec(f,0)
    for t in range(0,360,theta):
        Irot = iainterot( Iab, t, DIRECTION )
        y = iaunion( y, iasupgen(f, Irot))

    return y




# -*- encoding: utf-8 -*-
# Module iadrawv

from numpy import *
from string import upper

def iadrawv(f, data, value, GEOM):

    GEOM  = GEOM.upper()
    data  = array(data)
    y     = array(f)
    value = array(value)
    lin, col = data[1,:], data[0,:]
    i = lin*f.shape[1] + col
    i = i.astype(int32)
    if len(f.shape) == 1: f = f[newaxis,:]
    if value.shape == (): value = value + zeros(lin.shape)
    if len(lin) != len(value):
        print 'Number of points must match n. of colors.'
        return None
    if GEOM == 'POINT':
        ravel(y)[i] = value
    elif GEOM == 'LINE':
        for k in range(len(value)-1):
            delta = 1.*(lin[k+1]-lin[k])/(1e-10 + col[k+1]-col[k])
            if abs(delta) <= 1:
                if col[k] < col[k+1]: x_ = arange(col[k],col[k+1]+1)
                else                : x_ = arange(col[k+1],col[k]+1)
                y_ = floor(delta*(x_-col[k]) + lin[k] + 0.5)
            else:
                if lin[k] < lin[k+1]: y_ = arange(lin[k],lin[k+1]+1)
                else                : y_ = arange(lin[k+1],lin[k]+1)
                x_ = floor((y_-lin[k])/delta + col[k] + 0.5)
            i_ = y_*f.shape[1] + x_; i_ = i_.astype(int32)
            ravel(y)[i_]=value[k]
    elif GEOM == 'RECT':
        for k in range(data.shape[1]):
            d = data[:,k]
            x0,y0,x1,y1 = d[1],d[0],d[3],d[2]
            y[x0:x1,y0]   = value[k]
            y[x0:x1,y1]   = value[k]
            y[x0,y0:y1]   = value[k]
            y[x1,y0:y1+1] = value[k]
    elif GEOM == 'FRECT':
        for k in range(data.shape[1]):
            d = data[:,k]
            x0,y0,x1,y1 = d[1],d[0],d[3],d[2]
            y[x0:x1+1,y0:y1+1] = value[k]
    else:
        print "GEOM should be 'POINT', 'LINE', 'RECT', or 'FRECT'."
    return y




# -*- encoding: utf-8 -*-
# Module iainpos

from numpy import *

from iasecross import iasecross
def iainpos(f, g, bc=iasecross()):
    from iaisbinary import iaisbinary
    from iagray import iagray
    from ianeg import ianeg
    from iadatatype import iadatatype
    from ialimits import ialimits
    from iasuprec import iasuprec
    from iaintersec import iaintersec
    from iaunion import iaunion

    assert iaisbinary(f),'First parameter must be binary image'
    fg = iagray( ianeg(f), iadatatype(g))
    k1 = ialimits(g)[1] - 1
    y = iasuprec(fg, iaintersec( iaunion(g, 1), k1, fg), bc)

    return y




# -*- encoding: utf-8 -*-
# Module iablob

import numpy as np
from string import upper
import ia870

def iablob(fr, measurement, option="image"):

    measurement = upper(measurement)
    option      = upper(option)
    if len(fr.shape) == 1: fr = fr[newaxis,:]
    n = fr.max()
    if   option      == 'DATA':        y = []
    elif measurement == 'CENTROID':    y = np.zeros(fr.shape,bool)
    elif measurement == 'BOUNDINGBOX': y = np.zeros(fr.shape,bool)
    elif measurement == 'PERIMETER':   y = np.zeros(fr.shape,np.int32)
    elif measurement == 'AREA':        y = np.zeros(fr.shape,np.int32)
    else:                              y = np.zeros(fr.shape)

    if measurement == 'AREA':
        area = np.bincount(fr.ravel())
        if option == 'DATA':
            y = area[1::]
        else:
            for i in xrange(1,n+1):
               y[fr==i] = area[i]

    elif measurement == 'CENTROID':
        for i in xrange(1,n+1):
            aux  = (fr==i)
            xind,yind = np.nonzero(aux)
            area = len(xind)
            centroid = [xind.sum()/area,yind.sum()/area]
            if option == 'DATA': y.append([centroid[1],centroid[0]])
            else               : y[centroid[0],centroid[1]] = 1

    elif measurement == 'BOUNDINGBOX':
        for i in xrange(1,n+1):
            aux = (fr==i)
            col, = np.nonzero(aux.any(0))
            row, = np.nonzero(aux.any(1))
            if option == 'DATA': y.append([col[0],row[0],col[-1],row[-1]])
            else:
                y[row[0]:row[-1],col[0] ] = 1
                y[row[0]:row[-1],col[-1]] = 1
                y[row[0], col[0]:col[-1]] = 1
                y[row[-1],col[0]:col[-1]] = 1

    elif measurement == 'PERIMETER':
        Bc = ia870.iasecross()
        for i in xrange(1,n+1):
           aux = fr == i
           grad = aux - ia870.iaero(aux,Bc)
           if option == 'DATA': y.append(grad.sum())
           else:
               y[aux] = grad.sum()

    elif measurement == 'CIRCULARITY':
        Bc = ia870.iasecross()
        area = np.bincount(fr.ravel())
        perim = []
        for i in xrange(1,n+1):
           aux = fr == i
           grad = aux - ia870.iaero(aux,Bc)
           perim.append(grad.sum())
           if option != 'DATA':
               y[aux] = 4*np.pi*area[i]/(perim[i-1]**2)
        if option == 'DATA':
            perim = np.array(perim)
            y = 4*np.pi*area[1::]/(perim**2)

    elif measurement == 'ASPECTRATIO':
        for i in xrange(1,n+1):
            aux = (fr==i)
            col, = np.nonzero(aux.any(0))
            row, = np.nonzero(aux.any(1))
            if option == 'DATA': y.append(1.*(min(col[-1]-col[0],row[-1]-row[0])+1)/(max(col[-1]-col[0],row[-1]-row[0])+1))
            else:
                y[aux] = 1.*(min(col[-1]-col[0],row[-1]-row[0])+1)/(max(col[-1]-col[0],row[-1]-row[0])+1)

    elif measurement == 'COMPACTNESS':
        for i in xrange(1,n+1):
            aux  = (fr==i)
            xind,yind = np.nonzero(aux)
            area = len(xind)
            centroid = [xind.sum()/area,yind.sum()/area]
            m20 = ((1.*xind-centroid[0])**2).sum()/area
            m02 = ((1.*yind-centroid[1])**2).sum()/area
            compactness = area/(2*np.pi*(m20+m02))
            if option == 'DATA': y.append(compactness)
            else               : y[aux] = compactness

    elif measurement == 'ECCENTRICITY':
        for i in xrange(1,n+1):
            aux  = (fr==i)
            xind,yind = np.nonzero(aux)
            area = len(xind)
            centroid = [xind.sum()/area,yind.sum()/area]
            m11 = ((1.*xind-centroid[0])*(yind-centroid[1])).sum()/area
            m20 = ((1.*xind-centroid[0])**2).sum()/area
            m02 = ((1.*yind-centroid[1])**2).sum()/area
            eccentricity = sqrt((m20-m02)**2+4*m11**2)/(m20+m02)
            if option == 'DATA': y.append(eccentricity)
            else               : y[aux] = eccentricity

    else:
        print "Measurement option should be 'AREA','CENTROID', 'BOUNDINGBOX', 'PERIMETER', "
        print "'ASPECTRATIO', 'CIRCULARITY', 'COMPACTNESS' or 'ECCENTRICITY'."

    return np.array(y)




# -*- encoding: utf-8 -*-
# Module iaregmin

from numpy import *
from iasecross import iasecross

def iaregmin(f, Bc=iasecross(), option="binary"):
    from iahmin import iahmin
    from iaaddm import iaaddm
    from iasubm import iasubm
    from iabinary import iabinary
    from iasuprec import iasuprec
    from iaunion import iaunion
    from iathreshad import iathreshad

    if option != "binary":
        raise Exception("iaregmin accepts only binary option")

    return iabinary(iasubm(iahmin(f,1,Bc), f))




# -*- encoding: utf-8 -*-
# Module ipdp

from numpy import array, ravel

# constants
# neighbourhood 2D
N4 = array([[-1,0],[0,1],[1,0],[0,-1]])
N8 = array([[-1,0],[0,1],[1,0],[0,-1],[-1,-1],[-1,1],[1,1],[1,-1]])

# neighbourhood 3D
N6 = array([[-1,0,0],[0,1,0],[1,0,0],[0,-1,0],[0,0,1],[0,0,-1]])


# values
inf = 1e400
BORDER = inf

def isBorder(v):
    """ Tests if the value is a border value """
    return v == BORDER

# common classes for the algorithms
class wsImage():
    """ Class for storing the images """

    def __init__(self, array):
        """ Constructor for storing the N-D input image """
        self.input = array
        self.work = None
        self.label = None
        self.output = None

    def begin(self, offsets):
        """ Prepare the image for processing """
        from numpy import zeros, ravel

        if len(self.input.shape) != offsets.shape[1]:
            raise Exception("Image shape does not fit offsets dimensions")

        # initialise the neighbour
        self.neighbour = wsNeighbour(offsets)
        # pad the input image
        self.work = self._pad()
        # store the padded shape
        self.workshape = self.work.shape
        # ravel the padded image (make 1-D)
        self.work = ravel(self.work)
        # make a zeroed copy of it
        self.label = zeros(self.work.shape)
        # initialise the output
        self.output = None
        # initialise the shape of the image
        self.neighbour.setImageShape(self.workshape)
        self.neighbour.setImage(self.work)
        # initialise the domain object
        D = wsDomain(self.work.size)
        D.setImage(self.work)

        # returns the neighbourhood relation, the working image, the label
        # image and the domain of the image
        return self.neighbour.N, self.work, self.label, D


    def _pad(self):
        from numpy import zeros
        """ Pads the N-D image with the BORDER constant as necessary for
        containing all the offsets from the list
        """
        # generate the newshape by iterating through the original and
        # adding the extra space needed at each dimension
        newshape = tuple(map(lambda orig, d: orig + (d-1), self.input.shape, self.neighbour.shape))
        # generate the slicing list
        slicing = map(lambda orig, d: slice((d-1)/2, (d-1)/2 + orig), self.input.shape, self.neighbour.shape)
        # create the padded image
        workimage = zeros(newshape)
        workimage[:] = BORDER
        workimage[slicing] = self.input
        return workimage

    def _crop(self):
        return self.crop(self.label)

    def crop(self, x):
        """ Reshape and crops the N-D label image to the original size (same as self.input) """
        from numpy import reshape
        # generate the slicing list
        slicing = map(lambda orig, d: slice((d-1)/2, (d-1)/2 + orig), self.input.shape, self.neighbour.shape)
        # reshape the label image to the original shape
        temp = reshape(x, self.workshape)
        # crop the temp image
        return temp[slicing]

    def end(self):
        return self._crop().astype('int32')

    def makeWorkCopy(self, default=0):
        """ Make a copy of the work image filled with the value and type of parameter default """
        copied = self.work.copy()
        copied = copied.astype(type(default))
        copied.fill(default)
        return copied

    def getCoordinate(self, p):
        """ Get the coordinates in (x,y) form for index p"""
        return (p / self.workshape[1], p % self.workshape[1])


class wsNeighbour():
    """ Class for neighbourhood processing """

    def __init__(self, offsets):
        """ Constructor for the list of offsets in N-D (neighbours)
            offsets must be a m x N matrix, where m is the number of
            offsets (neighbours) and N is the dimensions of the image"""
        self.offsets = array(offsets)
        self._shape()
        self.s = None

    def _shape(self):
        """ Calculates the shape of the offsets """
        N = self.offsets.shape[1]
        self.shape = []
        for i in range(N):
            dmax = max(self.offsets[:,i])
            dmin = min(self.offsets[:,i])
            if abs(dmax) != abs(dmin):
                raise Exception("Offsets must be symmetrical")
            d = dmax - dmin + 1
            # make the dimension always odd
            if d % 2 == 0:
                d += 1
            self.shape.append(d)
        self.shape = tuple(self.shape)

    def setImageShape(self, imshape):
        """ Set the image shape and calculates the offsets in 1-D """
        self.s = imshape
        if len(self.s) != self.offsets.shape[1]:
            raise Exception("Image shape does not fit offsets dimensions")

        # calculate the offsets in 1-D
        # the process occurs like this:
        # each offset is multiplied by the multiplication of the values of
        # the next components of the shape of the image and summed:
        # example:
        # shape: (3, 10, 15)
        # offset: [1, 1, 2]
        # offset in 1-D: (1 * 10 * 15) + (1 * 15) + (2)
        #
        # of course, the offsets must follow the order of the shape
        # (Nth-D, ..., 3rd-D, 2nd-D, 1st-D), that is usually
        # (time, channel, row, column) or in grayscale images (time, row, column)
        # or simple 2-D images (row, column)

        # LONG VERSION
        # self.roffsets = []
        # for offset in self.offsets:
            # roffset = 0
            # for i in range(len(offset)):
                # n = offset[i]
                # roffset += n * reduce(lambda x,y: x*y, self.s[(i+1):], 1)
            # self.roffsets.append(roffset)

        # SHORT VERSION (using map and reduce - I like this one better)
        self.roffsets = map(
            lambda offset: sum(
                map(lambda n, i:
                        n * reduce(lambda x,y: x*y, self.s[(i+1):], 1),
                    offset, range(len(offset))
                    )
                ),
            self.offsets)

    def setImage(self, im):
        """ Set the working image to query for border values on neighbourhood calculation """
        self.im = im

    def N(self, pixel):
        """ Returns the list of indexes of neighbours of pixel in 1-D """
        if not self.s:
            raise Exception("Set the image shape first!")

        # calculate the coordinates of the neighbours based on the offsets in 1-D
        n = map(lambda c: c + pixel, self.roffsets)
        r = list()

        for i in n:
            if isBorder(self.im[i]):
                continue

            r.append(i)

        return r

    def query(self, c):
        """ Lookup on the roffsets for the index of the offset c """
        for i in range(len(self.roffsets)):
            if self.roffsets[i] == c:
                return i
        return -1

    def addOffset(self, p, index):
        """ Adds the offset of the desired index to the value p """
        if index < 0 or index >= len(self.roffsets):
            return None
        else:
            c = self.roffsets[index]
            return p + c

class wsDomain:
    def __init__(self, length):
        self.inner = xrange(length)
        self.count = 0

    def next(self):
        if self.count >= len(self.inner):
            self.count = 0
            raise StopIteration

        while isBorder(self.im[self.inner[self.count]]):
            self.count += 1
            if self.count >= len(self.inner):
                self.count = 0
                raise StopIteration

        c = self.count
        self.count = c + 1
        return self.inner[c]

    def __getitem__(self, item):
        return self.inner[item]

    def __iter__(self):
        return self

    def __len__(self):
        return len(self.inner)

    def setImage(self, im):
        self.im = im

class wsQueue():
    """ Simple queue class for abstracting list methods """

    def __init__(self):
        self.array = list()

    def push(self, a):
        """ Pushes an element to the end of the queue """
        self.array.append(a)

    def pop(self):
        """ Pops the first element of the queue """
        return self.array.pop(0)

    def empty(self):
        """ Returns whether the queue is empty or not """
        return len(self.array) == 0

    def clear(self):
        """ Clear the list """
        self.array = list()

class wsStack():
    """ Simple stack class for abstracting list methods """

    def __init__(self):
        self.array = list()

    def push(self, a):
        """ Pushes an element to the top of the stack """
        self.array.append(a)

    def pop(self):
        """ Pops the top element of the stack """
        return self.array.pop()

    def empty(self):
        """ Returns whether the stack is empty or not """
        return len(self.array) == 0

    def clear(self):
        """ Clear the list """
        self.array = list()

class wsHeapQueue():
    """ Priority queue class for abstracting list methods with FIFO policy """

    def __init__(self):
        self.queue = dict()

    def push(self, a, c):
        """ Pushes an element to the queue """
        if self.queue.has_key(c):
            self.queue[c].append(a)
        else:
            self.queue[c] = [a]

    def pop(self):
        """ Pops the first element of the queue """
        key = min(self.queue.keys())
        element = self.queue[key].pop(0)
        if len(self.queue[key]) == 0:
            self.queue.pop(key)
        return element

    def empty(self):
        """ Returns whether the queue is empty or not """
        return len(self.queue) == 0

    def clear(self):
        """ Clear the queue """
        self.queue = dict()

    def remove(self, a, c):
        """ Remove the element a at cost c """
        self.queue[c].remove(a)

    def contains(self, a, c):
        """ Verifies if the queue contains element a at cost c """
        for x in self.queue[c]:
            if x == a:
                return True
        return False

class wsRandHeapQueue():
    """ Priority queue class for abstracting list methods without FIFO policy """

    class wsPixel():

        def __init__(self, p, l):
            self.p = p
            self.l = l

        def __lt__(self, other):
            return self.l < other.l

        def __le__(self, other):
            return self.l <= other.l

        def __eq__(self, other):
            return self.l == other.l and self.p == other.p

        def __ne__(self, other):
            return not (self.l == other.l and self.p == other.p)

        def __gt__(self, other):
            return self.l > other.l

        def __ge__(self, other):
            return self.l >= other.l

    def __init__(self):
        self.queue = []

    def push(self, a, c):
        from heapq import heappush
        """ Pushes an element to the queue """
        px = wsRandHeapQueue.wsPixel(a, c)
        heappush(self.queue, px)

    def pop(self):
        from heapq import heappop
        """ Pops the first element of the queue """
        px = heappop(self.queue)
        return px.p

    def empty(self):
        """ Returns whether the queue is empty or not """
        return len(self.queue) == 0

    def clear(self):
        """ Clear the queue """
        self.queue = []

    def remove(self, a, c):
        from heapq import heapify
        """ Remove the element a at cost c """
        self.queue.remove(wsRandHeapQueue.wsPixel(a,c))
        heapify(self.queue)

    def contains(self, a, c):
        """ Verifies if the queue contains element a at cost c """
        for px in self.queue:
            if px.p == a and px.l == c:
                return True
        return False


def findMinima(im, N, D):

    UNVISITED = 0
    PENDING = 1
    VISITED = 2

    work = im.copy()
    work[:] = UNVISITED

    qPending = wsQueue()
    qMinima = wsQueue()

    minima = list()

    count = 1

    for p in D:
        if work[p] != UNVISITED:
            continue

        qPending.push(p)
        work[p] = PENDING

        is_minima = True
        qMinima.clear()

        while not qPending.empty():

            q = qPending.pop()
            qMinima.push(q)
            work[q] = VISITED

            for u in N(q):

                if im[u] == im[q] and work[u] == UNVISITED:
                    work[u] = PENDING
                    qPending.push(u)
                elif im[u] < im[q]:
                    is_minima = False # keep flooding the plateau



        if is_minima:
            count += 1
            seed = list()
            while not qMinima.empty():
                q = qMinima.pop()
                seed.append(q)
            minima.append(seed)

    return minima

def lowerComplete(im, offsets):

    # initialise variables
    ws = wsImage(im)
    N, im, lc, D = ws.begin(offsets)

    FICTITIOUS_PIXEL = -1
    queue = wsQueue()

    lc[:] = BORDER

    for p in D:

        lc[p] = 0
        for q in N(p):
            if im[q] < im[p]:
                queue.push(p)
                lc[p] = -1
                break

    cur_dist = 1

    queue.push(FICTITIOUS_PIXEL)

    while not queue.empty():
        p = queue.pop()
        if p == FICTITIOUS_PIXEL:
            if not queue.empty():
                queue.push(FICTITIOUS_PIXEL)
                cur_dist += 1
        else:
            lc[p] = cur_dist
            for q in N(p):

                if im[q] == im[p] and lc[q] == 0:
                    queue.push(q)
                    lc[q] = -1

    for p in D:

        if lc[p] == 0:
            lc[p] = cur_dist * im[p]
        else:
            lc[p] = cur_dist * im[p] + lc[p] - 1

    return ws.end()

def se2offset(se):
    from numpy import array
    from ia870 import iaseshow

    se = iaseshow(se, 'NORMAL')

    offset = []
    for i in range(se.shape[0]):
        for j in range(se.shape[1]):
            if se[i,j] == True and (i-se.shape[0]/2 != 0 or j-se.shape[1]/2 != 0):
                offset.append([i-se.shape[0]/2,j-se.shape[1]/2])

    return array(offset)




# -*- encoding: utf-8 -*-
# Module iaasfrec

from numpy import *
from string import upper
from iasecross import iasecross

def iaasfrec(f, SEQ="OC", b=iasecross(), bc=iasecross(), n=1):
    from iasesum import iasesum
    from iacloserec import iacloserec
    from iaopenrec import iaopenrec

    SEQ = upper(SEQ)
    y = f
    if SEQ == 'OC':
        for i in range(1,n+1):
            nb = iasesum(b,i)
            y = iacloserec(y,nb,bc)
            y = iaopenrec(y,nb,bc)
    elif SEQ == 'CO':
        for i in range(1,n+1):
            nb = iasesum(b,i)
            y = iaopenrec(y,nb,bc)
            y = iacloserec(y,nb,bc)
    else:
        assert 0,'Only accepts OC or CO for SEQ parameter'

    return y




# -*- encoding: utf-8 -*-
# Module iaintershow

from numpy import *
from string import join

def iaintershow(Iab):
    from iaseunion import iaseunion
    from iaintersec import iaintersec


    assert (type(Iab) is tuple) and (len(Iab) == 2),'not proper fortmat of hit-or-miss template'
    A,Bc = Iab
    S = iaseunion(A,Bc)
    Z = iaintersec(S,0)
    n = product(S.shape)
    one  = reshape(array(n*'1','c'),S.shape)
    zero = reshape(array(n*'0','c'),S.shape)
    x    = reshape(array(n*'.','c'),S.shape)
    saux = choose( S + iaseunion(Z,A), ( x, zero, one))
    s = ''
    for i in range(saux.shape[0]):
        s=s+(join(list(saux[i]))+' \n')

    return s




# -*- encoding: utf-8 -*-
# Module ialimits

from numpy import *

def ialimits(f):


    code = f.dtype
    if   code == bool:   y=array([0,1],'bool')
    elif code == uint8:  y=array([0,255],'uint8')
    elif code == uint16: y=array([0,(2**16)-1],'uint16')
    elif code == int32:  y=array([-((2**31)-1),(2**31)-1],'int32')
    elif code == int64:  y=array([-((2**63)-1), (2**63)-1],'int64')
    elif code == float64:y=array([-Inf,Inf],'float64')
    else:
        assert 0,'ialimits: Does not accept this typecode:%s' % code
    return y




# -*- encoding: utf-8 -*-
# Module iaseline

from numpy import *

def iaseline(l=3, theta=0):
    from iaset2mat import iaset2mat
    from iabinary import iabinary

    theta = pi*theta/180
    if abs(tan(theta)) <= 1:
        s  = sign(cos(theta))
        x0 = arange(0, l * cos(theta)-(s*0.5),s)
        x1 = floor(x0 * tan(theta) + 0.5)
    else:
        s  = sign(sin(theta))
        x1 = arange(0, l * sin(theta) - (s*0.5),s)
        x0 = floor(x1 / tan(theta) + 0.5)
    x = transpose(array([x1,x0],int32))
    B = iaset2mat((x,iabinary(ones((x.shape[1],1)))))
    return B




# -*- encoding: utf-8 -*-
# Module iaset2mat

from numpy import *

def iaset2mat(A):
    from iabinary import iabinary
    from ialimits import ialimits

    if len(A) == 2:
        x, v = A
        v = asarray(v)
    elif len(A) == 1:
        x = A[0]
        v = ones((len(x),),bool)
    else:
        raise TypeError, 'Argument must be a tuple of length 1 or 2'
    if len(x) == 0:  return array([0]).astype(v.dtype)
    if len(x.shape) == 1: x = x[newaxis,:]
    dh, dw = abs(x).max(0)
    h,w = (2*dh) + 1, (2*dw) + 1
    M=ones((h, w)) * ialimits(v)[0]
    offset = x[:,0] * w + x[:,1] + (dh*w + dw)
    M.flat[offset] = v
    M = M.astype(v.dtype)

    return M




# -*- encoding: utf-8 -*-
# Module iaNlut

import numpy as np
def iaNlut(s,offset):
    '''Precompute array of neighbors. Optimized by broadcast.

    s - image shape
    offset - offset matrix, 2 columns (dh,dw) by n. of neighbors rows
    '''
    H,W = s
    n = H*W
    hi = np.arange(H).reshape(-1,1)
    wi = np.arange(W).reshape(1,-1)
    hoff = offset[:,0]
    woff = offset[:,1]
    h = hi + hoff.reshape(-1,1,1)
    w = wi + woff.reshape(-1,1,1)
    h[(h<0) | (h>=H)] = n
    w[(w<0) | (w>=W)] = n
    Nlut = np.clip(h * W + w,0,n)
    return Nlut.reshape(offset.shape[0],-1).transpose()




# -*- encoding: utf-8 -*-
# Module iagshow

from numpy import *

def iagshow(X, X1=None, X2=None, X3=None, X4=None, X5=None, X6=None):
    from iaisbinary import iaisbinary
    from iagray import iagray
    from iaunion import iaunion
    from iaintersec import iaintersec
    from ianeg import ianeg
    from iaconcat import iaconcat

    if iaisbinary(X): X = iagray(X,'uint8')
    r = X
    g = X
    b = X
    if X1 is not None: # red 1 0 0
      assert iaisbinary(X1),'X1 must be binary overlay'
      x1 = iagray(X1,'uint8')
      r = iaunion(r,x1)
      g = iaintersec(g,ianeg(x1))
      b = iaintersec(b,ianeg(x1))
    if X2 is not None: # green 0 1 0
      assert iaisbinary(X2),'X2 must be binary overlay'
      x2 = iagray(X2,'uint8')
      r = iaintersec(r,ianeg(x2))
      g = iaunion(g,x2)
      b = iaintersec(b,ianeg(x2))
    if X3 is not None: # blue 0 0 1
      assert iaisbinary(X3),'X3 must be binary overlay'
      x3 = iagray(X3,'uint8')
      r = iaintersec(r,ianeg(x3))
      g = iaintersec(g,ianeg(x3))
      b = iaunion(b,x3)
    if X4 is not None: # magenta 1 0 1
      assert iaisbinary(X4),'X4 must be binary overlay'
      x4 = iagray(X4,'uint8')
      r = iaunion(r,x4)
      g = iaintersec(g,ianeg(x4))
      b = iaunion(b,x4)
    if X5 is not None: # yellow 1 1 0
      assert iaisbinary(X5),'X5 must be binary overlay'
      x5 = iagray(X5,'uint8')
      r = iaunion(r,x5)
      g = iaunion(g,x5)
      b = iaintersec(b,ianeg(x5))
    if X6 is not None: # cyan 0 1 1
      assert iaisbinary(X6),'X6 must be binary overlay'
      x6 = iagray(X6,'uint8')
      r = iaintersec(r,ianeg(x6))
      g = iaunion(g,x6)
      b = iaunion(b,x6)
    return iaconcat('d',r,g,b)
    return Y




# -*- encoding: utf-8 -*-
# Module iastats

from numpy import *
from string import upper

def iastats(f, measurement):

    measurement = upper(measurement)
    if measurement == 'MAX':
        y = max(ravel(f))
    elif measurement == 'MIN':
        y = min(ravel(f))
    elif measurement == 'MEAN':
        y = mean(ravel(f))
    elif measurement == 'MEDIAN':
        y = median(ravel(f))
    elif measurement == 'STD':
        y = std(ravel(f),ddof=1)
    elif measurement == 'STD1':
        y = std(ravel(f))
    elif measurement == 'SUM':
        y = sum(ravel(f))
    else:
        assert 0,'Not a valid measurement'


    return y




# -*- encoding: utf-8 -*-
# Module iaedgeoff

from numpy import *
from iasecross import iasecross

def iaedgeoff(f, Bc=iasecross()):
    from iaframe import iaframe
    from iasubm import iasubm
    from iainfrec import iainfrec

    edge = iaframe(f)
    return iasubm( f, iainfrec(edge, f, Bc))




# -*- encoding: utf-8 -*-
# Module iaseintersec

from numpy import *

def iaseintersec(B1, B2):
    from ialimits import ialimits
    from ia870 import iamat2set, iaset2mat

    assert B1.dtype == B2.dtype, \
      'iaseintersec: Cannot have different datatypes: \
      %s and %s' % (str(B1.dtype), str(B2.dtype))
    type1 = B1.dtype
    #if len(B1) == 0: return B2
    if len(B1.shape) == 1: B1 = B1[newaxis,:]
    if len(B2.shape) == 1: B2 = B2[newaxis,:]
    if B1.shape <> B2.shape:
        inf = ialimits(B1)[0]
        h1,w1 = B1.shape
        h2,w2 = B2.shape
        H,W = max(h1,h2),max(w1,w2)
        Hc,Wc = (H-1)/2,(W-1)/2    # center
        BB1,BB2 = asarray(B1),asarray(B2)
        B1, B2  = inf * ones((H,W)), inf *ones((H,W))
        dh1s , dh1e = (h1-1)/2 , (h1-1)/2 + (h1+1)%2 # deal with even and odd dimensions
        dw1s , dw1e = (w1-1)/2 , (w1-1)/2 + (w1+1)%2
        dh2s , dh2e = (h2-1)/2 , (h2-1)/2 + (h2+1)%2
        dw2s , dw2e = (w2-1)/2 , (w2-1)/2 + (w2+1)%2
        B1[ Hc-dh1s : Hc+dh1e+1  ,  Wc-dw1s : Wc+dw1e+1 ] = BB1
        B2[ Hc-dh2s : Hc+dh2e+1  ,  Wc-dw2s : Wc+dw2e+1 ] = BB2
    B = minimum(B1,B2).astype(type1)
    i,v = iamat2set(B)
    B = iaset2mat((i,v))
    return B

